from .utils import TestCase


class Test(TestCase):

    def test_for_loop_underscore_variable(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    for _ in [1, 4]:\n'
            '        print(_)\n',
            '  File "", line 3\n'
            '            print(_)\n'
            '                  ^\n'
            "CompileError: undefined variable '_'\n")

    def test_iterate_over_tuple(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    for i in (5, 1):\n'
            '        print(i)\n',
            '  File "", line 2\n'
            '        for i in (5, 1):\n'
            '                 ^\n'
            "CompileError: iteration over tuples not allowed\n")

    def test_iterate_over_enumerate_not_tuple(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    values: [u32] = [3, 8]\n'
            '    for i in enumerate(values):\n'
            '        print(i)\n',
            '  File "", line 3\n'
            '        for i in enumerate(values):\n'
            '            ^\n'
            "CompileError: can only unpack enumerate into two variables, got 1\n")

    def test_iterate_over_slice_with_different_types_1(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    for i in slice([1], 1, u16(2)):\n'
            '        print(i)\n',
            '  File "", line 2\n'
            '        for i in slice([1], 1, u16(2)):\n'
            '                               ^\n'
            "CompileError: types 'u16' and 'i64' differs\n")

    def test_iterate_over_slice_with_different_types_2(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    for i in slice(range(4), 1, 2, i8(-1)):\n'
            '        print(i)\n',
            '  File "", line 2\n'
            '        for i in slice(range(4), 1, 2, i8(-1)):\n'
            '                                       ^\n'
            "CompileError: types 'i8' and 'i64' differs\n")

    def test_iterate_over_slice_no_params(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    for i in slice(range(2)):\n'
            '        print(i)\n',
            '  File "", line 2\n'
            '        for i in slice(range(2)):\n'
            '                 ^\n'
            "CompileError: expected 2 to 4 parameters, got 1\n")

    def test_iterate_over_range_with_different_types_1(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    for i in range(i8(1), u16(2)):\n'
            '        print(i)\n',
            '  File "", line 2\n'
            '        for i in range(i8(1), u16(2)):\n'
            '                       ^\n'
            "CompileError: types 'i8' and 'i64' differs\n")

    def test_iterate_over_range_with_different_types_2(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    for i in range(1, i8(2), 2):\n'
            '        print(i)\n',
            '  File "", line 2\n'
            '        for i in range(1, i8(2), 2):\n'
            '                          ^\n'
            "CompileError: types 'i8' and 'i64' differs\n")

    def test_iterate_over_range_with_too_many_parameters(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    for i in range(1, 2, 2, 2):\n'
            '        print(i)\n',
            '  File "", line 2\n'
            '        for i in range(1, 2, 2, 2):\n'
            '                 ^\n'
            "CompileError: expected 1 to 3 parameters, got 4\n")

    def test_iterate_over_enumerate_no_parameters(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    for i, j in enumerate():\n'
            '        print(i)\n',
            '  File "", line 2\n'
            '        for i, j in enumerate():\n'
            '                    ^\n'
            "CompileError: expected 1 or 2 parameters, got 0\n")

    def test_iterate_over_zip_wrong_unpack(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    for i in zip(range(2), range(2)):\n'
            '        print(i)\n',
            '  File "", line 2\n'
            '        for i in zip(range(2), range(2)):\n'
            '            ^\n'
            "CompileError: cannot unpack 2 values into 1\n")

    def test_iterate_over_reversed_no_parameter(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    for i in reversed():\n'
            '        print(i)\n',
            '  File "", line 2\n'
            '        for i in reversed():\n'
            '                 ^\n'
            "CompileError: expected 1 parameter, got 0\n")

    def test_bare_integer_in_for(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    for i in "":\n'
            '        1\n',
            '  File "", line 3\n'
            '            1\n'
            '            ^\n'
            "CompileError: bare integer\n")

    def test_variable_defined_in_for_not_be_used_after(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    l: [bool] = []\n'
            '    for _ in l:\n'
            '        v = 1\n'
            '    print(v)\n',
            '  File "", line 5\n'
            '        print(v)\n'
            '              ^\n'
            "CompileError: undefined variable 'v'\n")

    def test_fail_to_redefine_method_call_variable_in_for_loop_for_now_1(self):
        self.assert_transpile_raises(
            'class Foo:\n'
            '    def foo(self) -> [string]:\n'
            '        return []\n'
            'def foo(a: bool):\n'
            '    for a in Foo().foo():\n'
            '        print(a)\n',
            '  File "", line 5\n'
            '        for a in Foo().foo():\n'
            '            ^\n'
            "CompileError: redefining variable 'a'\n")

    def test_fail_to_redefine_method_call_variable_in_for_loop_for_now_2(self):
        self.assert_transpile_raises(
            'class Foo:\n'
            '    def foo(self) -> [(string, u8)]:\n'
            '        return []\n'
            'def foo(b: bool):\n'
            '    for a, b in Foo().foo():\n'
            '        print(a, b)\n',
            '  File "", line 5\n'
            '        for a, b in Foo().foo():\n'
            '               ^\n'
            "CompileError: redefining variable 'b'\n")

    def test_fail_slice_dict(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    for k, v in slice({1: 2}):\n'
            '        print(k, v)\n',
            '  File "", line 2\n'
            '        for k, v in slice({1: 2}):\n'
            '                          ^\n'
            "CompileError: unsupported iterator type {i64: i64}\n")



import shutil

from ..utils import DOWNLOAD_DIRECTORY
from ..utils import add_url_argument
from ..utils import add_verbose_argument
from ..utils import download_dependencies
from ..utils import read_package_configuration
from ..utils import setup_build


def do_fetch(_parser, args, _mys_config):
    config = read_package_configuration()
    shutil.rmtree(DOWNLOAD_DIRECTORY, ignore_errors=True)
    setup_build()
    download_dependencies(config, args.url)


def add_subparser(subparsers):
    subparser = subparsers.add_parser(
        'fetch',
        description='Download and extract all dependencies.')
    add_verbose_argument(subparser)
    add_url_argument(subparser)
    subparser.set_defaults(func=do_fetch)



from collections import defaultdict

from .utils import CompileError
from .utils import is_primitive_type
from .utils import is_snake_case
from .utils import split_dict_mys_type


class State:

    def __init__(self, variables, raises):
        self.variables = variables
        self.raises = raises


class SpecializedFunction:

    def __init__(self, function, call_module_name, call_node):
        self.function = function
        self.first_call_module_name = call_module_name
        self.first_call_node = call_node

    def __str__(self):
        return f'SpecializedFunction(function={self.function})'


class SpecializedClass:

    def __init__(self, definitions, call_module_name, call_node):
        self.definitions = definitions
        self.first_call_module_name = call_module_name
        self.first_call_node = call_node

    def __str__(self):
        return f'SpecializedClass(definitions={self.definitions})'


class Traceback:

    def __init__(self, source_lines):
        self.source_lines = source_lines
        self.entries = []
        self.index = -1
        self.name = None

    def add(self, lineno):
        if lineno - 1 < len(self.source_lines):
            code = self.source_lines[lineno - 1].strip()
        else:
            code = "known bug"

        code = code.replace('\\', '').replace('"', '\\"')
        self.entries.append((self.name, lineno, code))

        return len(self.entries) - 1

    def enter(self, name):
        self.name = name

        return '    __MYS_TRACEBACK_ENTER();'

    def exit(self):
        return '    __MYS_TRACEBACK_EXIT();'

    def set(self, lineno):
        index = self.add(lineno)

        return f'    __MYS_TRACEBACK_SET({index});'


class Context:
    """The context keeps track of defined functions, classes, traits,
    enums and variables in the current scope. Ot also provides other
    services, like unique numbers and full name convertions.

    """

    def __init__(self,
                 module_levels,
                 specialized_functions,
                 specialized_classes,
                 source_lines):
        self.name = '.'.join(module_levels)
        self._stack = [[]]
        self.local_variables = {}
        self._global_variables = {}
        self._classes = {}
        self._traits = {}
        self._functions = {}
        self._enums = {}
        self.return_mys_type = None
        self.mys_type = None
        self.unique_count = 0
        self.constants = {}
        self._name_to_full_name ={}
        self.specialized_functions = specialized_functions
        self.specialized_classes = specialized_classes
        self.comprehensions = []
        self._raises = [False]
        self.source_lines = source_lines
        self.class_name = None
        self.method_comprehensions = defaultdict(list)
        self.traceback = Traceback(source_lines)
        self.package = module_levels[0]

    def unique_number(self):
        self.unique_count += 1

        return self.unique_count

    def unique(self, name):
        return f'__{name}_{self.unique_number()}'

    def make_full_name(self, name):
        """Returns the fully qualified name (full_name) of given function,
        class, trait, enum or global variable name.

        """

        return self._name_to_full_name.get(name)

    def define_local_variable(self, name, mys_type, node):
        if self.is_local_variable_defined(name):
            raise CompileError(f"redefining variable '{name}'", node)

        if not is_snake_case(name):
            raise CompileError("local variable names must be snake case", node)

        if name in self._name_to_full_name:
            raise CompileError(f"redefining '{name}'", node)

        self.local_variables[name] = mys_type
        self._stack[-1].append(name)

    def is_local_variable_defined(self, name):
        """Returns true if given short local variable name is defined.

        """

        return name in self.local_variables

    def get_local_variable_type(self, name):
        return self.local_variables[name]

    def define_global_variable(self, name, full_name, mys_type, _node):
        self._global_variables[name] = mys_type
        self._name_to_full_name[name] = full_name

    def is_global_variable_defined(self, name):
        """Returns true if given name short global variable name is defined.

        """

        return name in self._global_variables

    def get_global_variable_type(self, name):
        return self._global_variables[name]

    def make_full_name_this_module(self, name):
        """Returns the fully qualified name (full_name) of given name as
        defined in the current module.

        """

        return f'{self.name}.{name}'

    def define_class(self, name, full_name, definitions):
        self._name_to_full_name[name] = full_name
        self._classes[full_name] = definitions

    def is_class_defined(self, name):
        """Returns true if given type is a class. Accepts both short names and
        fully qualified names.

        """

        if not isinstance(name, str):
            return False

        full_name = self._name_to_full_name.get(name, name)

        return full_name in self._classes

    def get_class_definitions(self, name):
        full_name = self._name_to_full_name.get(name, name)

        return self._classes[full_name]

    def define_trait(self, name, full_name, definitions):
        self._name_to_full_name[name] = full_name
        self._traits[full_name] = definitions

    def is_trait_defined(self, name):
        """Returns true if given type is a trait. Accepts both short names and
        fully qualified names.

        """

        if not isinstance(name, str):
            return False

        full_name = self._name_to_full_name.get(name, name)

        return full_name in self._traits

    def get_trait_definitions(self, name):
        full_name = self._name_to_full_name.get(name, name)

        return self._traits[full_name]

    def define_function(self, name, full_name, definitions):
        self._name_to_full_name[name] = full_name
        self._functions[full_name] = definitions

    def is_function_defined(self, full_name):
        """Returns true if given fully qualified function name is defined.

        """

        return full_name in self._functions

    def get_functions(self, full_name):
        return self._functions[full_name]

    def define_enum(self, name, full_name, definitions):
        self._name_to_full_name[name] = full_name
        self._enums[full_name] = definitions

    def is_enum_defined(self, name):
        """Returns true if given type is an enum. Accepts both short names and
        fully qualified names.

        """

        if not isinstance(name, str):
            return False

        full_name = self._name_to_full_name.get(name, name)

        return full_name in self._enums

    def get_enum_type(self, name):
        return self.get_enum_definitions(name).type

    def get_enum_definitions(self, name):
        full_name = self._name_to_full_name.get(name, name)

        return self._enums[full_name]

    def is_class_or_trait_defined(self, full_name):
        if self.is_class_defined(full_name):
            return True
        elif self.is_trait_defined(full_name):
            return True

        return False

    def is_type_defined(self, mys_type):
        if isinstance(mys_type, tuple):
            for item_mys_type in mys_type:
                if not self.is_type_defined(item_mys_type):
                    return False
        elif isinstance(mys_type, list):
            for item_mys_type in mys_type:
                if not self.is_type_defined(item_mys_type):
                    return False
        elif isinstance(mys_type, dict):
            key_mys_type, value_mys_type = split_dict_mys_type(mys_type)

            if not self.is_type_defined(key_mys_type):
                return False

            if not self.is_type_defined(value_mys_type):
                return False
        elif isinstance(mys_type, set):
            for item_mys_type in mys_type:
                if not self.is_type_defined(item_mys_type):
                    return False
        elif self.is_class_or_trait_defined(mys_type):
            return True
        elif self.is_enum_defined(mys_type):
            return True
        elif is_primitive_type(mys_type):
            return True
        elif mys_type in ('string', 'bytes', 'regex', 'regexmatch'):
            return True
        elif mys_type is None:
            return True
        else:
            return False

        return True

    def define_specialized_function(self, full_name, function, call_node):
        self.specialized_functions[full_name] = SpecializedFunction(
            function,
            self.name,
            call_node)

    def is_specialized_function_defined(self, full_name):
        return full_name in self.specialized_functions

    def get_specialized_function(self, full_name):
        return self.specialized_functions[full_name].function

    def define_specialized_class(self, full_name, definitions, call_node):
        self.specialized_classes[full_name] = SpecializedClass(
            definitions,
            self.name,
            call_node)

    def is_specialized_class_defined(self, full_name):
        return full_name in self.specialized_classes

    def get_specialized_class(self, full_name):
        return self.specialized_classes[full_name].definitions

    def push(self):
        self._stack.append([])
        self._raises.append(False)

    def pop(self):
        result = {}

        for name in self._stack[-1]:
            result[name] = self.local_variables.pop(name)

        self._stack.pop()

        return State(result, self._raises.pop())

    def set_always_raises(self, value):
        self._raises[-1] = value

    def __str__(self):
        result = ['Context:']
        result.append('  Local variables:')

        for name, value in self.local_variables.items():
            result.append(f'    {name}: {value}')

        result.append('  Global variables:')

        for name, value in self._global_variables.items():
            result.append(f'    {name}: {value}')

        result.append('  Classes:')

        for name, value in self._classes.items():
            result.append(f'    {name}: {value}')

        return '\n'.join(result)



# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/nedbat/coveragepy/blob/master/NOTICE.txt

"""Code coverage measurement for Python.

Ned Batchelder
https://nedbatchelder.com/code/coverage

"""

from .control import Coverage
from .sqldata import CoverageData



from ..parser import ast
from .utils import CompileError


class ImportsVisitor(ast.NodeVisitor):
    """Raises an error if not all imports are at the beginning of the
    file.

    """

    def __init__(self):
        self._other_found = False

    def visit_Module(self, node):
        for item in node.body:
            self.visit(item)

    def visit_Import(self, node):
        raise CompileError("only 'from <module> import ...' is allowed", node)

    def visit_ImportFrom(self, node):
        if self._other_found:
            raise CompileError("imports must be at the beginning of the file", node)

    def generic_visit(self, node):
        self._other_found = True



def _find_root_modules(modules):
    """Returns a list of modules that no other module imports from.

    """

    module_imports = {name: 0 for name in modules}

    for neighbors in modules.values():
        for neighbor in neighbors:
            module_imports[neighbor] += 1

    return sorted([name for name, count in module_imports.items() if count == 0])


class _Importer:

    def __init__(self, modules):
        self.modules = modules
        self.seen = set()

    def import_module(self, name):
        """Recursivelly find all modules given module imports and the module
        iteslf in the order they shall be initialised.

        """

        # Break circular imports.
        if name in self.seen:
            return []

        self.seen.add(name)
        ordered = []

        for module in self.modules[name]:
            ordered += self.import_module(module)

        ordered.append(name)

        return ordered


def resolve_import_order(modules):
    """Returns a list of all modules ordered by import order suitable for
    initialization. Best effort for cyclic imports.

    """

    root_modules = _find_root_modules(modules)
    ordered = []

    for root_module in root_modules:
        ordered += _Importer(modules).import_module(root_module)

    for module in sorted(modules):
        if module not in ordered:
            ordered += _Importer(modules).import_module(module)

    return ordered



from setuptools import setup

setup(name='{name}',
      version='{version}',
      description='{description}',
      long_description=open('README.rst', 'r').read(),
      author={author},
      author_email={author_email},
      install_requires={dependencies})



from mys.transpiler import Source
from mys.transpiler import transpile

from .utils import TestCase


def transpile_source(source, filename='', module_hpp='', has_main=False):
    return transpile([Source(source,
                             filename=filename,
                             module_hpp=module_hpp,
                             has_main=has_main)])[0][2]

class Test(TestCase):

    def test_if_string(self):
        source = transpile_source('def foo(v: string):\n'
                                  '    if v == "a":\n'
                                  '        print(v)\n'
                                  '    elif v == "b":\n'
                                  '        print(v)\n'
                                  '    elif "c" == v:\n'
                                  '        print(v)\n')

        self.assert_in(
            'static const mys::String __constant_1 = mys::String("a");', source)
        self.assert_in(
            'static const mys::String __constant_2 = mys::String("b");', source)
        self.assert_in(
            'static const mys::String __constant_3 = mys::String("c");', source)
        self.assert_in('if (mys::Bool(v == __constant_1)) {', source)
        self.assert_in('if (mys::Bool(v == __constant_2)) {', source)
        self.assert_in('if (mys::Bool(__constant_3 == v)) {', source)

    def test_if_in_list(self):
        source = transpile_source('def foo(v: i32):\n'
                                  '    if v in [1, 2]:\n'
                                  '        print(v)\n')

        self.assert_in(
            'static const SharedList<i32> __constant_1 = mys::make_shared',
            source)
        self.assert_in('if (mys::Bool(contains(v, __constant_1))) {', source)

    def test_if_mix_of_types(self):
        source = transpile_source('def foo(v: (bool, [string], (u8, i8))):\n'
                                  '    if v != (True, [], (1, -5)):\n'
                                  '        print(v)\n')

        self.assert_in(
            'static const SharedTuple<mys::Bool, SharedList<mys::String>, '
            'SharedTuple<u8, i8>> '
            '__constant_1 = mys::make_shared',
            source)
        self.assert_in('if (mys::Bool(v != __constant_1)) {', source)

    def test_if_mix_of_types_not_constant(self):
        source = transpile_source('def foo(v: (bool, [string], (u8, i8)), k: bool):\n'
                                  '    if v != (k, [], (1, -5)):\n'
                                  '        print(v)\n')

        self.assertNotIn('static const SharedTuple', source)
        self.assert_in('if (mys::Bool(v != mys::make_shared<', source)

    def test_if_primitive_types_not_constant(self):
        source = transpile_source('def foo(v: i64):\n'
                                  '    if v == 1:\n'
                                  '        print(v)\n')

        self.assertNotIn('static const i64', source)
        self.assert_in('if (mys::Bool(v == 1)', source)

    def test_match_string(self):
        source = transpile_source('def foo(v: string):\n'
                                  '    res = 0\n'
                                  '    match v:\n'
                                  '        case "123":\n'
                                  '            res = 1\n'
                                  '        case "hi":\n'
                                  '            res = 2\n'
                                  '    print(res)\n')

        self.assert_in(
            'static const mys::String __constant_2 = mys::String("123");',
            source)
        self.assert_in(
            'static const mys::String __constant_3 = mys::String("hi");',
            source)
        self.assert_in('if (__subject_1 == __constant_2) {', source)
        self.assert_in('if (__subject_1 == __constant_3) {', source)



import re
import shutil
from html.parser import HTMLParser
from unittest.mock import patch

import mys.cli

from .utils import Path
from .utils import TestCase
from .utils import read_file
from .utils import remove_build_directory


class HTMLText(HTMLParser):

    def __init__(self):
        super().__init__()
        self.text = ''

    def handle_data(self, data):
        self.text += data


class Test(TestCase):

    def test_doc(self):
        name = 'test_doc'
        remove_build_directory(name)

        shutil.copytree('tests/files/doc', f'tests/build/{name}')

        with Path(f'tests/build/{name}'):
            with patch('sys.argv', ['mys', '-d', 'doc', '-v']):
                mys.cli.main()

            self.assert_file_exists('build/doc/html/index.html')

            html_text = HTMLText()
            html_text.feed(read_file('build/doc/html/index.html'))

            with open('index.txt', 'w') as fout:
                mo = re.search(
                    r'(The doc package in the Mys programming language.*)© Copyright',
                    html_text.text,
                    re.DOTALL)
                fout.write(mo.group(1).strip())

            self.assert_files_equal('index.txt', '../../files/doc/index.txt')



import glob

from ....parser import ast
from ...utils import Spinner
from ...utils import read_package_configuration
from .comments_finder import CommentsFinder
from .comments_reader import CommentsReader
from .source_styler import SourceStyler
from .utils import get_function_or_class_node_start
from .utils import get_source


def style_files():
    source_styler = SourceStyler()

    for src_path in glob.glob('src/**/*.mys', recursive=True):
        with open(src_path, 'r') as fin:
            source = fin.read()

        tree = ast.parse(source, src_path)
        source_lines = source.split('\n')
        finder = CommentsFinder(source_lines)
        finder.visit(tree)
        # from pprint import pprint
        # pprint(finder.comments)
        styled_source = source_styler.style(source_lines, tree, finder.comments)

        if styled_source != source:
            with open(src_path, 'w') as fout:
                fout.write(styled_source)


def do_style(_parser, _args, _mys_config):
    read_package_configuration()

    with Spinner(text="Styling source code"):
        style_files()


def add_subparser(subparsers):
    subparser = subparsers.add_parser(
        'style',
        description=(
            'Modify all source code files in the package follow the Mys style '
            'guidelines.'))
    subparser.set_defaults(func=do_style)



import glob
import multiprocessing
import os
import re
import shutil
import subprocess
import sys
import tarfile
import time

import requests
import toml
import yaspin
from colors import blue
from colors import cyan
from colors import green
from colors import red
from colors import strip_color
from colors import yellow
from humanfriendly import format_timespan

from ..coverage import Coverage
from ..coverage import CoverageData
from ..transpiler.utils import is_snake_case
from ..version import __version__

MYS_DIR = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))

DOWNLOAD_DIRECTORY = 'build/dependencies'

BULB = yellow('💡', style='bold')
INFO = blue('🛈 ', style='bold')
ERROR = red('❌️', style='bold')

OPTIMIZE = {
    'speed': '3',
    'size': 's',
    'debug': '0'
}

TRANSPILE_OPTIONS_FMT = '-n {package_name} -p {package_path} {flags}'

COPY_HPP_AND_CPP_FMT = '''\
{dst}: {src}
\tmkdir -p $(dir $@)
\tcp $< $@
'''

COPY_ASSET_FMT = '''\
{dst}: {src}
\tmkdir -p $(dir $@)
\tcp $< $@
'''

class BuildConfig:

    def __init__(self,
                 debug,
                 verbose,
                 optimize,
                 debug_symbols,
                 no_ccache,
                 coverage,
                 unsafe,
                 jobs,
                 url):
        self.debug = debug
        self.verbose = verbose
        self.optimize = optimize
        self.debug_symbols = debug_symbols
        self.no_ccache = no_ccache
        self.coverage = coverage
        self.unsafe = unsafe
        self.jobs = jobs
        self.url = url

def create_file(path, data):
    with open(path, 'w') as fout:
        fout.write(data)


def read_template_file(path):
    with open(os.path.join(MYS_DIR, 'cli/templates', path)) as fin:
        return fin.read()


def default_jobs():
    return max(1, multiprocessing.cpu_count() - 1)


def duration_start():
    return time.time()


def duration_stop(start_time):
    end_time = time.time()
    duration = format_timespan(end_time - start_time)

    return f' ({duration})'


def box_print(lines, icon, width=None):
    if width is None:
        width = 0

        for line in lines:
            width = max(width, len(strip_color(line)))

    print(f'┌{"─" * (width - 3)} {icon} ─┐')

    for line in lines:
        padding = width - len(strip_color(line))
        print(f'│ {line}{" " * padding} │')

    print(f'└{"─" * (width + 2)}┘')


SPINNER = [
    ' ⠋', ' ⠙', ' ⠹', ' ⠸', ' ⠼', ' ⠴', ' ⠦', ' ⠧', ' ⠇', ' ⠏'
]


class Spinner(yaspin.api.Yaspin):

    def __init__(self, text):
        super().__init__(yaspin.Spinner(SPINNER, 80), text=text, color='yellow')
        self._start_time = duration_start()

    def __exit__(self, exc_type, exc_val, traceback):
        duration = duration_stop(self._start_time)

        if exc_type is None:
            self.write(green(' ✔ ') + self.text + duration)
        else:
            self.write(red(' ✘ ') + self.text + duration)

        return super().__exit__(exc_type, exc_val, traceback)


def run_with_spinner(command, message, env=None):
    output = ''

    try:
        with Spinner(text=message):
            result = subprocess.run(command,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT,
                                    encoding='utf-8',
                                    close_fds=False,
                                    env=env)
            output = result.stdout
            result.check_returncode()
    except Exception:
        lines = []

        for line in output.splitlines():
            if 'make: *** ' in line:
                continue

            lines.append(line)

        raise Exception('\n'.join(lines).rstrip())

    return output


def run(command, message, verbose, env=None):
    if verbose:
        start_time = duration_start()

        try:
            print('Command:', ' '.join(command))

            with subprocess.Popen(command,
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.STDOUT,
                                  encoding='utf-8',
                                  close_fds=False,
                                  env=env) as proc:
                output = []

                while proc.poll() is None:
                    text = proc.stdout.readline()
                    print(text, end="")
                    output.append(text)

                print(proc.stdout.read(), end="")

                if proc.returncode != 0:
                    raise Exception(f'command failed with {proc.returncode}')

            output = ''.join(output)
            print(green(' ✔ ') + message + duration_stop(start_time))
        except Exception:
            print(red(' ✘ ') + message + duration_stop(start_time))
            raise
    else:
        output = run_with_spinner(command, message, env)

    return output


def create_file_from_template(path, directory, **kwargs):
    template = read_template_file(os.path.join(directory, path))
    create_file(path, template.format(**kwargs))


def create_file_from_template_2(path, template_path, **kwargs):
    template = read_template_file(template_path)
    create_file(path, template.format(**kwargs))


class Author:

    def __init__(self, name, email):
        self.name = name
        self.email = email


class PackageConfig:

    def __init__(self, path):
        self.authors = []
        self.config = self.load_package_configuration(path)

    def load_package_configuration(self, path):
        with open(os.path.join(path, 'package.toml')) as fin:
            config = toml.loads(fin.read())

        package = config.get('package')

        if package is None:
            raise Exception("'[package]' not found in package.toml.")

        for name in ['name', 'version', 'authors']:
            if name not in package:
                raise Exception(f"'[package].{name}' not found in package.toml.")

        if not is_snake_case(package['name']):
            raise Exception(
                f"package name must be snake case, got '{package['name']}'")

        for author in package['authors']:
            mo = re.match(r'^([^<]+)<([^>]+)>$', author)

            if not mo:
                raise Exception(f"Bad author '{author}'.")

            self.authors.append(Author(mo.group(1).strip(), mo.group(2).strip()))

        if 'description' not in package:
            package['description'] = ''

        dependencies = {
            'fiber': {'path': os.path.join(MYS_DIR, 'lib/packages/fiber')}
        }

        if 'dependencies' in config:
            dependencies.update(config['dependencies'])

        config['dependencies'] = dependencies

        if 'c-dependencies' not in config:
            config['c-dependencies'] = {}

        return config

    def __getitem__(self, key):
        return self.config[key]


def setup_build():
    os.makedirs(DOWNLOAD_DIRECTORY, exist_ok=True)


def rename_one_matching(pattern, to):
    paths = glob.glob(pattern)

    if len(paths) != 1:
        raise Exception(
            f'{len(paths)} paths are matching when expecting exactly one to match')

    os.rename(paths[0], to)


def prepare_download_dependency_from_registry(name, version):
    archive = f'{name}-{version}.tar.gz'
    archive_path = f'{DOWNLOAD_DIRECTORY}/{archive}'

    if os.path.exists(archive_path):
        return None
    else:
        return (name, version, archive, archive_path)


def extract_dependency(name, version, archive, archive_path):
    if version == 'latest':
        rename_one_matching(os.path.join(DOWNLOAD_DIRECTORY, f'{name}-*.tar.gz'),
                            archive_path)

    with Spinner(text=f"Extracting {archive}"):
        with tarfile.open(archive_path) as fin:
            fin.extractall(DOWNLOAD_DIRECTORY)

    if version == 'latest':
        rename_one_matching(os.path.join(DOWNLOAD_DIRECTORY, f'{name}-*/'),
                            os.path.join(DOWNLOAD_DIRECTORY, f'{name}-latest'))


def download_dependencies(config, url):
    packages = []

    for name, info in config['dependencies'].items():
        if isinstance(info, str):
            package = prepare_download_dependency_from_registry(name, info)

            if package is not None:
                packages.append(package)

    if not packages:
        return

    with Spinner(text="Downloading dependencies"):
        for _, _, archive, archive_path in packages:
            response = requests.get(f'{url}/package/{archive}')

            if response.status_code != 200:
                print(response.text)

                raise Exception('Package download failed.')

            with open(archive_path, 'wb') as fout:
                fout.write(response.content)

    for name, version, archive, archive_path in packages:
        extract_dependency(name, version, archive, archive_path)


def read_package_configuration():
    try:
        with Spinner('Reading package configuration'):
            return PackageConfig('.')
    except FileNotFoundError:
        box_print([
            'Current directory does not contain a Mys package (package.toml does',
            'not exist).',
            '',
            'Please enter a Mys package directory, and try again.',
            '',
            f'You can create a new package with {cyan("mys new <name>")}.'],
                  BULB)

        raise Exception()


def find_package_sources(package_name, path, ignore_main=False):
    srcs_mys = []
    srcs_hpp = []
    srcs_cpp = []
    oldpath = os.getcwd()
    os.chdir(os.path.join(path, 'src'))

    try:
        for src in glob.glob('**/*.mys', recursive=True):
            if ignore_main and src == 'main.mys':
                continue

            srcs_mys.append((package_name, path, src, os.path.join(path, 'src', src)))

        for src in glob.glob('**/*.hpp', recursive=True):
            srcs_hpp.append((package_name, path, src, os.path.join(path, 'src', src)))

        for src in glob.glob('**/*.cpp', recursive=True):
            srcs_cpp.append((package_name, path, src, os.path.join(path, 'src', src)))
    finally:
        os.chdir(oldpath)

    return srcs_mys, srcs_hpp, srcs_cpp


def dependency_path(dependency_name, config):
    for package_name, info in config['dependencies'].items():
        if package_name == dependency_name:
            if isinstance(info, str):
                return f'build/dependencies/{package_name}-{info}/'
            elif 'path' in info:
                return info['path']
            else:
                raise Exception('Bad dependency format.')

    raise Exception(f'Bad dependency {dependency_name}.')


def find_dependency_sources(config):
    srcs_mys = []
    srcs_hpp = []
    srcs_cpp = []
    packages_paths = set()

    for package_name in config['dependencies']:
        path = dependency_path(package_name, config)
        srcs = find_package_sources(package_name, path, ignore_main=True)
        srcs_mys += srcs[0]
        srcs_hpp += srcs[1]
        srcs_cpp += srcs[2]
        packages_paths.add(path)

    return srcs_mys, srcs_hpp, srcs_cpp, sorted(packages_paths)


def find_assets(config):
    assets = []
    paths = [(config['package']['name'], '.')]

    for package_name in config['dependencies']:
        paths.append((package_name, dependency_path(package_name, config)))

    for package_name, path in paths:
        oldpath = os.getcwd()
        assets_path = os.path.join(path, 'assets')

        if not os.path.exists(assets_path):
            continue

        os.chdir(assets_path)

        try:
            for asset in glob.glob('**/*', recursive=True):
                if os.path.isfile(asset):
                    assets.append((package_name, assets_path, asset))
        finally:
            os.chdir(oldpath)

    return assets


class PkgConfigFlags:

    def __init__(self):
        self.flags = []

    def add(self, output):
        # ToDo: Spaces in path.
        for flag in output.strip().split():
            if flag not in self.flags:
                self.flags.append(flag)

    def __str__(self):
        return ' '.join(self.flags)


def find_c_dependencies_flags(packages_paths, verbose, cflags, libs):
    if shutil.which('mys-config'):
        pkg_config = 'mys-config'
    elif shutil.which('pkg-config'):
        pkg_config = 'pkg-config'
    else:
        raise Exception('mys-config nor pkg-config found')

    for path in packages_paths:
        config = PackageConfig(path)

        for library_name in config['c-dependencies']:
            output = run([pkg_config, library_name, '--cflags'],
                         f'Getting compiler flags {library_name}',
                         verbose)
            cflags.add(output)
            output = run([pkg_config, library_name, '--libs'],
                         f'Getting linker flags for {library_name}',
                         verbose)
            libs.add(output)


def create_makefile(config, build_config):
    combo = build_config.optimize

    if build_config.coverage:
        combo += '-coverage'

    if build_config.unsafe:
        combo += '-unsafe'

    build_dir = f'build/{combo}'

    os.makedirs(f'{build_dir}/cpp', exist_ok=True)
    srcs_mys, srcs_hpp, srcs_cpp = find_package_sources(
        config['package']['name'],
        '.')
    cflags = PkgConfigFlags()
    libs = PkgConfigFlags()
    find_c_dependencies_flags(['.'], build_config.verbose, cflags, libs)

    if not srcs_mys:
        box_print(["'src/' is empty. Please create one or more .mys-files."], ERROR)

        raise Exception()

    srcs = find_dependency_sources(config)
    srcs_mys += srcs[0]
    srcs_hpp += srcs[1]
    srcs_cpp += srcs[2]
    packages_paths = srcs[3]

    transpile_options = []
    transpile_srcs = []
    transpile_srcs_paths = []
    copy_hpp_and_cpp = []
    objs = []
    is_application = False
    transpiled_cpp = []
    hpps = []
    find_c_dependencies_flags(packages_paths, build_config.verbose, cflags, libs)

    if build_config.debug_symbols:
        cflags.flags.append('-g')

    assets = find_assets(config)

    for package_name, package_path, src, _path in srcs_mys:
        flags = []

        if package_name != config['package']['name']:
            flags.append('-s yes')
        else:
            flags.append('-s no')

        if src == 'main.mys':
            is_application = True
            flags.append('-m yes')
        else:
            flags.append('-m no')

        flags = ' '.join(flags)

        module_path = f'$(BUILD)/cpp/src/{package_name}/{src}'
        transpile_options.append(
            TRANSPILE_OPTIONS_FMT.format(package_name=package_name,
                                         package_path=package_path,
                                         flags=flags))

        transpile_srcs.append(src)
        transpile_srcs_paths.append(os.path.join(package_path, 'src', src))
        objs.append(f'OBJ += {module_path}.$(OBJ_SUFFIX)')
        transpiled_cpp.append(f'SRC += {module_path}.cpp')

    for package_name, package_path, src, _path in srcs_hpp:
        src_path = os.path.join(package_path, 'src', src)
        module_path = f'$(BUILD)/cpp/src/{package_name}/{src}'
        copy_hpp_and_cpp.append(COPY_HPP_AND_CPP_FMT.format(src=src_path,
                                                            dst=module_path))
        hpps.append(module_path)

    for package_name, package_path, src, _path in srcs_cpp:
        src_path = os.path.join(package_path, 'src', src)
        module_path = f'$(BUILD)/cpp/src/{package_name}/{src}'
        copy_hpp_and_cpp.append(COPY_HPP_AND_CPP_FMT.format(src=src_path,
                                                            dst=module_path))
        objs.append(f'OBJ += {module_path}.o')
        transpiled_cpp.append(f'SRC += {module_path}.cpp')

    copy_assets = []
    assets_targets = []

    for package_name, assets_path, asset in assets:
        target = os.path.join(f'$(EXE)-assets/{package_name}', asset)
        copy_assets.append(
            COPY_ASSET_FMT.format(src=os.path.join(assets_path, asset),
                                  dst=target))
        assets_targets.append(target)

    if is_application:
        all_deps = '$(EXE)'
    else:
        all_deps = '$(OBJ)'

    if not build_config.no_ccache and shutil.which('ccache'):
        ccache = 'ccache '
    else:
        ccache = ''

    create_file_from_template_2(f'{build_dir}/Makefile',
                                'build/Makefile',
                                build=build_dir,
                                mys_dir=MYS_DIR,
                                mys=f'{sys.executable} -m mys',
                                ccache=ccache,
                                objs='\n'.join(objs),
                                optimize=OPTIMIZE[build_config.optimize],
                                transpile_options=' '.join(transpile_options),
                                transpile_srcs_paths=' '.join(transpile_srcs_paths),
                                transpile_srcs=' '.join(transpile_srcs),
                                hpps=' '.join(hpps),
                                copy_hpp_and_cpp='\n'.join(copy_hpp_and_cpp),
                                copy_assets='\n'.join(copy_assets),
                                assets=' '.join(assets_targets),
                                all_deps=all_deps,
                                package_name=config['package']['name'],
                                transpiled_cpp='\n'.join(transpiled_cpp),
                                cflags=cflags,
                                libs=libs)

    return is_application, build_dir


def build_prepare(build_config, config=None):
    if config is None:
        config = read_package_configuration()

    setup_build()
    download_dependencies(config, build_config.url)

    return create_makefile(config, build_config)


def build_app(build_config, is_application, build_dir):
    command = ['make', '-f', f'{build_dir}/Makefile', 'all']

    if os.getenv('MAKEFLAGS') is None:
        command += ['-j', str(build_config.jobs)]

    if build_config.debug:
        command += ['TRANSPILE_DEBUG=--debug']

    if not build_config.verbose:
        command += ['-s']

    if is_application:
        command += ['APPLICATION=yes']

    if build_config.coverage:
        command += ['COVERAGE=yes']

    if build_config.unsafe:
        command += ['UNSAFE=yes']

    if build_config.optimize == 'debug':
        command += ['TRACEBACK=yes']

    run(command, 'Building', build_config.verbose)


def add_verbose_argument(subparser):
    subparser.add_argument('-v', '--verbose',
                           action='store_true',
                           help='Verbose output.')


def add_jobs_argument(subparser):
    subparser.add_argument(
        '-j', '--jobs',
        type=int,
        default=default_jobs(),
        help='Maximum number of parallel jobs (default: %(default)s).')


def add_optimize_argument(subparser, default):
    subparser.add_argument(
        '-o', '--optimize',
        default=default,
        choices=['speed', 'size', 'debug'],
        help='Optimize the build for given level (default: %(default)s).')


def add_debug_symbols_argument(subparser):
    subparser.add_argument('-g', '--debug-symbols',
                           action='store_true',
                           help='Add debug symbols.')


def add_no_ccache_argument(subparser):
    subparser.add_argument('-n', '--no-ccache',
                           action='store_true',
                           help='Do not use ccache.')


def add_url_argument(subparser):
    subparser.add_argument('--url',
                           default='https://mys-lang.org',
                           help='Website URL (default: %(default)s).')


def add_coverage_argument(subparser):
    subparser.add_argument(
        '-c', '--coverage',
        action='store_true',
        help=('Instrument the code and create a coverage report when the '
              'application exits gracefully.'))


def add_unsafe_argument(subparser):
    subparser.add_argument(
        '-u', '--unsafe',
        action='store_true',
        help='Less runtime checks in favour of better performance.')


def _add_lines(coverage_data, path, linenos):
    coverage_data.add_lines(
        {path: {lineno: None for lineno in linenos}})


def _create_coverage_report(include):
    coverage_data = CoverageData()

    with open('.mys-coverage.txt', 'r') as fin:
        path = None
        linenos = []

        for line in fin:
            line = line.strip()

            if line.startswith('File:'):
                if path is not None:
                    _add_lines(coverage_data, path, linenos)

                path = os.path.abspath(line[6:])
                linenos = []
            else:
                lineno, count = line.split()

                if int(count) > 0:
                    linenos.append(int(lineno))

        if path is not None:
            _add_lines(coverage_data, path, linenos)

    coverage_data.write()

    cov = Coverage('.coverage', auto_data=True, include=include)
    cov.start()
    cov.stop()
    cov.html_report(directory='coverage/html')


def create_coverage_report(include=None):
    with Spinner('Creating code coverage report'):
        _create_coverage_report(include)

    path = os.path.abspath('coverage/html/index.html')
    print(f'Coverage report: file://{path}')



from .utils import TestCase
from .utils import build_and_test_module


class Test(TestCase):

    def test_loops(self):
        build_and_test_module('loops')

    def test_for_else_not_supported(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    for _ in range(3):\n'
            '        pass\n'
            '    else:\n'
            '        pass\n',
            '  File "", line 2\n'
            '        for _ in range(3):\n'
            '        ^\n'
            "CompileError: for else clause not supported\n")

    def test_while_else_not_supported(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    while False:\n'
            '        pass\n'
            '    else:\n'
            '        pass\n',
            '  File "", line 2\n'
            '        while False:\n'
            '        ^\n'
            "CompileError: while else clause not supported\n")



#!/usr/bin/env python

import re

from setuptools import Extension
from setuptools import find_packages
from setuptools import setup


def find_version():
    return re.search(r"^__version__ = '(.*)'$",
                     open('mys/version.py', 'r').read(),
                     re.MULTILINE).group(1)


setup(name='mys',
      version=find_version(),
      description='The Mys (/maɪs/) programming language.',
      long_description=open('README.rst', 'r').read(),
      author='Erik Moqvist',
      author_email='erik.moqvist@gmail.com',
      license='MIT',
      classifiers=[
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3',
      ],
      python_requires='>=3.8',
      keywords=['programming-language'],
      url='https://github.com/eerimoq/mys',
      packages=find_packages(exclude=['tests']),
      install_requires=[
          'ansicolors',
          'humanfriendly',
          'pygments',
          'toml',
          'yaspin',
          'sphinx',
          'sphinx_rtd_theme',
          'requests'
      ],
      include_package_data=True,
      ext_modules=[
          Extension('mys.parser._ast',
                    sources=[
                        'mys/parser/Python-ast.c',
                        'mys/parser/asdl.c',
                        'mys/parser/parser.c',
                        'mys/parser/peg_api.c',
                        'mys/parser/token.c',
                        'mys/parser/tokenizer.c',
                        'mys/parser/pegen.c',
                        'mys/parser/string_parser.c'
                    ])
      ],
      entry_points={
          'console_scripts': ['mys=mys.cli:main']
      })



import traceback

from pygments import highlight
from pygments.formatters import Terminal256Formatter
from pygments.lexer import RegexLexer
from pygments.lexer import bygroups
from pygments.lexer import using
from pygments.lexers import PythonLexer
from pygments.style import Style
from pygments.token import Generic
from pygments.token import Name
from pygments.token import Number
from pygments.token import Text

from ..parser import ast
from .class_transformer import ClassTransformer
from .coverage_transformer import CoverageTransformer
from .definitions import find_definitions
from .definitions import make_fully_qualified_names_module
from .header_visitor import HeaderVisitor
from .import_order import resolve_import_order
from .imports_visitor import ImportsVisitor
from .source_visitor import SourceVisitor
from .traits import ensure_that_trait_methods_are_implemented
from .utils import CompileError


class TranspilerError(Exception):
    pass


class TracebackLexer(RegexLexer):

    tokens = {
        'root': [
            (r'^(  File )("[^"]+")(, line )(\d+)(\n)',
             bygroups(Generic.Error, Name.Builtin, Generic.Error, Number, Text)),
            (r'^(\s+?)(\^)(\n)',
             bygroups(Text, Generic.Error, Text)),
            (r'^(    )(.+)(\n)',
             bygroups(Text, using(PythonLexer), Text)),
            (r'^([^:]+)(: )(.+)(\n)',
             bygroups(Generic.Escape, Text, Name, Text), '#pop')
        ]
    }


def format_location_source(source, lineno, offset):
    line = source.source_lines[lineno - 1]
    marker_line = ' ' * offset + '^'

    return (f'  File "{source.mys_path}", line {lineno}\n'
            f'    {line}\n'
            f'    {marker_line}\n')


def style_traceback(tb):
    return highlight(tb,
                     TracebackLexer(),
                     Terminal256Formatter(style='monokai'))


def transpile_file(tree,
                   source_lines,
                   filename,
                   module_hpp,
                   module_levels,
                   module_definitions,
                   definitions,
                   skip_tests,
                   has_main,
                   specialized_functions,
                   specialized_classes,
                   coverage_variables):
    namespace = 'mys::' + '::'.join(module_levels)
    source_visitor = SourceVisitor(namespace,
                                   module_levels,
                                   module_hpp,
                                   filename,
                                   source_lines,
                                   definitions,
                                   module_definitions,
                                   skip_tests,
                                   specialized_functions,
                                   specialized_classes,
                                   coverage_variables)
    source_visitor.visit(tree)
    header_visitor = HeaderVisitor(namespace,
                                   module_levels,
                                   module_hpp,
                                   source_lines,
                                   definitions,
                                   module_definitions,
                                   has_main,
                                   specialized_classes,
                                   source_visitor.method_comprehensions)
    header_visitor.visit(tree)

    return header_visitor, source_visitor


class Source:

    def __init__(self,
                 contents,
                 filename='',
                 module='',
                 mys_path='',
                 module_hpp='',
                 skip_tests=False,
                 hpp_path='',
                 cpp_path='',
                 has_main=False):
        self.contents = contents
        self.source_lines = contents.splitlines()
        self.filename = filename
        self.module = module
        self.module_levels = module.split('.')
        self.mys_path = mys_path
        self.module_hpp = module_hpp
        self.skip_tests = skip_tests
        self.hpp_path = hpp_path
        self.cpp_path = cpp_path
        self.has_main = has_main
        self.coverage_variables = {}

    def __str__(self):
        return '\n'.join([
            'Source:',
            f'  module: {self.module}',
            f'  filename: {self.filename}',
            f'  mys_path: {self.mys_path}',
            f'  skip_tests: {self.skip_tests}'
        ])

def transpile(sources, coverage=False):
    visitors = {}
    specialized_functions = {}
    specialized_classes = {}
    trees = []
    definitions = {}
    source_by_module = {source.module: source for source in sources}

    try:
        for source in sources:
            trees.append(ast.parse(source.contents, source.mys_path))
    except SyntaxError:
        lines = traceback.format_exc(0).splitlines()

        if 'Traceback (most recent call last):' in lines[0]:
            lines = lines[1:]

        raise Exception(style_traceback('\n'.join(lines)))

    source = None

    try:
        for source, tree in zip(sources, trees):
            ImportsVisitor().visit(tree)

        if coverage:
            for source, i in zip(sources, range(len(trees))):
                coverage_transformer = CoverageTransformer(source.contents)
                trees[i] = ast.fix_missing_locations(
                    coverage_transformer.visit(trees[i]))
                source.coverage_variables = coverage_transformer.variables()

        for source, i in zip(sources, range(len(trees))):
            trees[i] = ast.fix_missing_locations(ClassTransformer().visit(trees[i]))

        for source, tree in zip(sources, trees):
            definitions[source.module] = find_definitions(tree,
                                                          source.source_lines,
                                                          source.module_levels,
                                                          source.module)

        for source in sources:
            ensure_that_trait_methods_are_implemented(definitions[source.module],
                                                      definitions)

        for source in sources:
            make_fully_qualified_names_module(source.module,
                                              definitions[source.module])

        for source, tree in zip(sources, trees):
            header_visitor, source_visitor = transpile_file(
                tree,
                source.source_lines,
                source.mys_path,
                source.module_hpp,
                source.module_levels,
                definitions[source.module],
                definitions,
                source.skip_tests,
                source.has_main,
                specialized_functions,
                specialized_classes,
                source.coverage_variables)
            visitors[source.module] = (header_visitor, source_visitor)

        for name, function in specialized_functions.items():
            header_visitor, source_visitor = visitors['.'.join(name.split('.')[:-1])]
            header_visitor.visit_specialized_function(function.function)

            try:
                source_visitor.visit_specialized_function(function.function)
            except CompileError as e:
                raise TranspilerError(
                    style_traceback(
                        format_location_source(
                            source_by_module[function.first_call_module_name],
                            function.first_call_node.lineno,
                            function.first_call_node.col_offset)
                        + format_location_source(
                            source_by_module[function.function.module_name],
                            e.lineno,
                            e.offset)
                        + f'CompileError: {e.message}'))

        for name, klass in specialized_classes.items():
            header_visitor, source_visitor = visitors['.'.join(name.split('.')[:-1])]
            header_visitor.visit_specialized_class(name.split('.')[-1],
                                                   klass.definitions)

            try:
                source_visitor.visit_specialized_class(name.split('.')[-1],
                                                       klass.definitions)
            except CompileError as e:
                raise TranspilerError(
                    style_traceback(
                        format_location_source(
                            source_by_module[klass.first_call_module_name],
                            klass.first_call_node.lineno,
                            klass.first_call_node.col_offset)
                        + format_location_source(
                            source_by_module[klass.definitions.module_name],
                            e.lineno,
                            e.offset)
                        + f'CompileError: {e.message}'))

        module_imports = {}

        for source in sources:
            module_imports[source.module] = []

            for imports in definitions[source.module].imports.values():
                for imported_module, _ in imports:
                    if imported_module not in module_imports[source.module]:
                        module_imports[source.module].append(imported_module)

        ordered_modules = resolve_import_order(module_imports)
        last_source_visitor = visitors[ordered_modules[-1]][1]
        last_source_visitor.add_application_init(ordered_modules)
        last_source_visitor.add_application_exit(ordered_modules)

        return [
            (header_visitor.format_early_hpp(),
             header_visitor.format_hpp(),
             source_visitor.format_cpp())
            for header_visitor, source_visitor in visitors.values()
        ]
    except CompileError as e:
        raise TranspilerError(
            style_traceback(
                format_location_source(source, e.lineno, e.offset)
                + f'CompileError: {e.message}'))



from .utils import TestCase
from .utils import build_and_test_module
from .utils import transpile_source


class Test(TestCase):

    def test_char(self):
        build_and_test_module('char_')

    def test_bad_char_literal(self):
        self.assert_transpile_raises(
            'def foo():\n'
            "    print('foo')\n",
            '  File "", line 2\n'
            "        print('foo')\n"
            "              ^\n"
            'CompileError: bad character literal\n')

    def test_triple_quoted(self):
        with self.assertRaises(Exception) as cm:
            transpile_source('def foo():\n'
                             "    print('''f''')\n")

        self.assert_exception_string(
            cm,
            '  File "<string>", line 2\n'
            "    print('''f''')\n"
            "                 ^\n"
            'SyntaxError: characters cannot be triple quoted\n')

    def test_with_prefix(self):
        with self.assertRaises(Exception) as cm:
            transpile_source('def foo():\n'
                             "    print(b'f')\n")

        self.assert_exception_string(
            cm,
            '  File "<string>", line 2\n'
            "    print(b'f')\n"
            "              ^\n"
            'SyntaxError: characters cannot have a prefix\n')

    def test_concatenate(self):
        with self.assertRaises(Exception) as cm:
            transpile_source('def foo():\n'
                             "    print('f' 'o' 'o')\n")

        self.assert_exception_string(
            cm,
            '  File "<string>", line 2\n'
            "    print('f' 'o' 'o')\n"
            "                     ^\n"
            'SyntaxError: cannot concatenate characters\n')



from .utils import TestCase
from .utils import build_and_test_module


class Test(TestCase):

    def test_fstrings(self):
        build_and_test_module('fstrings')

    def test_invalid_format_spec(self):
        self.assert_transpile_raises(
            'def foo() -> string:\n'
            '    return f"{1:f}"\n',
            '  File "", line 2\n'
            '        return f"{1:f}"\n'
            '               ^\n'
            "CompileError: invalid integer format specifier 'f'\n")

    def test_invalid_format_spec_2(self):
        self.assert_transpile_raises(
            'def foo() -> string:\n'
            '    return f"{1:08d}"\n',
            '  File "", line 2\n'
            '        return f"{1:08d}"\n'
            '               ^\n'
            "CompileError: invalid integer format specifier '08d'\n")

    def test_float_with_format_spec(self):
        self.assert_transpile_raises(
            'def foo() -> string:\n'
            '    return f"{1.0:d}"\n',
            '  File "", line 2\n'
            '        return f"{1.0:d}"\n'
            '               ^\n'
            "CompileError: format specifiers are only allowed for integers\n")



# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/nedbat/coveragepy/blob/master/NOTICE.txt

"""Core control stuff for coverage.py."""

import contextlib
import os
import os.path

from .config import read_coverage_config
from .debug import DebugControl
from .files import abs_file
from .files import relative_filename
from .files import set_relative_directory
from .html import HtmlReporter
from .misc import DefaultValue
from .misc import ensure_dir_for_file
from .plugin import FileReporter
from .python import PythonFileReporter
from .results import Analysis
from .results import Numbers
from .sqldata import CoverageData


@contextlib.contextmanager
def override_config(cov, **kwargs):
    """Temporarily tweak the configuration of `cov`.

    The arguments are applied to `cov.config` with the `from_args` method.
    At the end of the with-statement, the old configuration is restored.
    """
    original_config = cov.config
    cov.config = cov.config.copy()
    try:
        cov.config.from_args(**kwargs)
        yield
    finally:
        cov.config = original_config


_DEFAULT_DATAFILE = DefaultValue("MISSING")

class Coverage:
    """Programmatic access to coverage.py.

    To use::

        from coverage import Coverage

        cov = Coverage()
        cov.start()
        #.. call your code ..
        cov.stop()
        cov.html_report(directory='covhtml')

    Note: in keeping with Python custom, names starting with underscore are
    not part of the public API. They might stop working at any point.  Please
    limit yourself to documented methods to avoid problems.

    """

    # The stack of started Coverage instances.
    _instances = []

    def __init__(
        self, data_file=_DEFAULT_DATAFILE, data_suffix=None, cover_pylib=None,
        auto_data=False, timid=None, branch=None, config_file=True,
        source=None, source_pkgs=None, omit=None, include=None, debug=None,
        check_preimported=False, context=None,
    ):  # pylint: disable=too-many-arguments
        """
        Many of these arguments duplicate and override values that can be
        provided in a configuration file.  Parameters that are missing here
        will use values from the config file.

        `data_file` is the base name of the data file to use. The config value
        defaults to ".coverage".  None can be provided to prevent writing a data
        file.  `data_suffix` is appended (with a dot) to `data_file` to create
        the final file name.  If `data_suffix` is simply True, then a suffix is
        created with the machine and process identity included.

        `cover_pylib` is a boolean determining whether Python code installed
        with the Python interpreter is measured.  This includes the Python
        standard library and any packages installed with the interpreter.

        If `auto_data` is true, then any existing data file will be read when
        coverage measurement starts, and data will be saved automatically when
        measurement stops.

        If `timid` is true, then a slower and simpler trace function will be
        used.  This is important for some environments where manipulation of
        tracing functions breaks the faster trace function.

        If `branch` is true, then branch coverage will be measured in addition
        to the usual statement coverage.

        `config_file` determines what configuration file to read:

            * If it is ".coveragerc", it is interpreted as if it were True,
              for backward compatibility.

            * If it is a string, it is the name of the file to read.  If the
              file can't be read, it is an error.

            * If it is True, then a few standard files names are tried
              (".coveragerc", "setup.cfg", "tox.ini").  It is not an error for
              these files to not be found.

            * If it is False, then no configuration file is read.

        `source` is a list of file paths or package names.  Only code located
        in the trees indicated by the file paths or package names will be
        measured.

        `source_pkgs` is a list of package names. It works the same as
        `source`, but can be used to name packages where the name can also be
        interpreted as a file path.

        `include` and `omit` are lists of file name patterns. Files that match
        `include` will be measured, files that match `omit` will not.  Each
        will also accept a single string argument.

        `debug` is a list of strings indicating what debugging information is
        desired.

        If `check_preimported` is true, then when coverage is started, the
        already-imported files will be checked to see if they should be
        measured by coverage.  Importing measured files before coverage is
        started can mean that code is missed.

        `context` is a string to use as the :ref:`static context
        <static_contexts>` label for collected data.

        .. versionadded:: 4.0
            The `concurrency` parameter.

        .. versionadded:: 4.2
            The `concurrency` parameter can now be a list of strings.

        .. versionadded:: 5.0
            The `check_preimported` and `context` parameters.

        .. versionadded:: 5.3
            The `source_pkgs` parameter.

        """
        # data_file=None means no disk file at all. data_file missing means
        # use the value from the config file.
        self._no_disk = data_file is None
        if data_file is _DEFAULT_DATAFILE:
            data_file = None

        # Build our configuration from a number of sources.
        self.config = read_coverage_config(
            config_file=config_file,
            data_file=data_file,
            cover_pylib=cover_pylib,
            timid=timid,
            branch=branch,
            parallel=None,
            source=source,
            source_pkgs=source_pkgs,
            run_omit=omit,
            run_include=include,
            debug=debug,
            report_omit=omit,
            report_include=include,
            concurrency=None,
            context=context,
            )

        # This is injectable by tests.
        self._debug_file = None

        self._auto_load = self._auto_save = auto_data
        self._data_suffix_specified = data_suffix

        # Is it ok for no data to be collected?
        self._warn_no_data = True
        self._warn_unimported_source = True
        self._warn_preimported_source = check_preimported
        self._no_warn_slugs = None

        # A record of all the warnings that have been issued.
        self._warnings = []

        # Other instance attributes, set later.
        self._data = self._collector = None
        self._inorout = None
        self._data_suffix = self._run_suffix = None
        self._exclude_re = None
        self._debug = None
        self._file_mapper = None

        # State machine variables:
        # Have we initialized everything?
        self._inited = False
        self._inited_for_start = False
        # Have we started collecting and not stopped it?
        self._started = False
        # Should we write the debug output?
        self._should_write_debug = True

    def _init(self):
        """Set all the initial state.

        This is called by the public methods to initialize state. This lets us
        construct a :class:`Coverage` object, then tweak its state before this
        function is called.

        """
        if self._inited:
            return

        self._inited = True

        # Create and configure the debugging controller. COVERAGE_DEBUG_FILE
        # is an environment variable, the name of a file to append debug logs
        # to.
        self._debug = DebugControl(self.config.debug, self._debug_file)

        # Multi-processing uses parallel for the subprocesses, so also use
        # it for the main process.
        self.config.parallel = False

        # _exclude_re is a dict that maps exclusion list names to compiled regexes.
        self._exclude_re = {}

        set_relative_directory()
        self._file_mapper = (relative_filename
                             if self.config.relative_files
                             else abs_file)

    def _post_init(self):
        """Stuff to do after everything is initialized."""
        if self._should_write_debug:
            self._should_write_debug = False
            self._write_startup_debug()

    def _write_startup_debug(self):
        """Write out debug info at startup if needed."""
        with self._debug.without_callers():
            if self._debug.should('config'):
                config_info = sorted(self.config.__dict__.items())
                config_info = [
                    (k, v) for k, v in config_info if not k.startswith('_')
                ]

    def load(self):
        """Load previously-collected coverage data from the data file."""
        self._init()
        if self._collector:
            self._collector.reset()
        should_skip = (self.config.parallel
                       and not os.path.exists(self.config.data_file))
        if not should_skip:
            self._init_data(suffix=None)
        self._post_init()
        if not should_skip:
            self._data.read()

    def _init_data(self, suffix):
        """Create a data file if we don't have one yet."""
        if self._data is None:
            # Create the data file.  We do this at construction time so that the
            # data file will be written into the directory where the process
            # started rather than wherever the process eventually chdir'd to.
            ensure_dir_for_file(self.config.data_file)
            self._data = CoverageData(
                basename=self.config.data_file,
                suffix=suffix,
                warn=None,
                debug=self._debug,
                no_disk=self._no_disk,
            )

    def start(self):
        """Start measuring code coverage.

        Coverage measurement only occurs in functions called after
        :meth:`start` is invoked.  Statements in the same scope as
        :meth:`start` won't be measured.

        Once you invoke :meth:`start`, you must also call :meth:`stop`
        eventually, or your process might not shut down cleanly.

        """

        if self._auto_load:
            self.load()

        self._started = True

    def stop(self):
        """Stop measuring code coverage."""
        self._started = False

    def get_data(self):
        """Get the collected data.

        Also warn about various problems collecting data.

        Returns a :class:`coverage.CoverageData`, the collected coverage data.

        .. versionadded:: 4.0

        """
        self._init()
        self._init_data(suffix=None)
        self._post_init()

        return self._data

    def _get_file_reporter(self, morf):
        """Get a FileReporter for a module or file name."""
        file_reporter = "python"

        if file_reporter == "python":
            file_reporter = PythonFileReporter(morf, self)

        return file_reporter

    def _analyze(self, it):
        """Analyze a single morf or code unit.

        Returns an `Analysis` object.

        """
        # All reporting comes through here, so do reporting initialization.
        self._init()
        Numbers.set_precision(self.config.precision)
        self._post_init()

        data = self.get_data()
        if not isinstance(it, FileReporter):
            it = self._get_file_reporter(it)

        return Analysis(data, it, self._file_mapper)


    def _get_file_reporters(self, morfs=None):
        """Get a list of FileReporters for a list of modules or file names.

        For each module or file name in `morfs`, find a FileReporter.  Return
        the list of FileReporters.

        If `morfs` is a single module or file name, this returns a list of one
        FileReporter.  If `morfs` is empty or None, then the list of all files
        measured is used to find the FileReporters.

        """
        if not morfs:
            morfs = self._data.measured_files()

        # Be sure we have a collection.
        if not isinstance(morfs, (list, tuple, set)):
            morfs = [morfs]

        file_reporters = [self._get_file_reporter(morf) for morf in morfs]
        return file_reporters

    def html_report(
        self, morfs=None, directory=None, ignore_errors=None,
        omit=None, include=None, extra_css=None, title=None,
        skip_covered=None, show_contexts=None, contexts=None,
        skip_empty=None, precision=None,
    ):
        """Generate an HTML report.

        The HTML is written to `directory`.  The file "index.html" is the
        overview starting point, with links to more detailed pages for
        individual modules.

        `extra_css` is a path to a file of other CSS to apply on the page.
        It will be copied into the HTML directory.

        `title` is a text string (not HTML) to use as the title of the HTML
        report.

        See :meth:`report` for other arguments.

        Returns a float, the total percentage covered.

        .. note::
            The HTML report files are generated incrementally based on the
            source files and coverage results. If you modify the report files,
            the changes will not be considered.  You should be careful about
            changing the files in the report folder.

        """
        with override_config(self,
                             ignore_errors=ignore_errors,
                             report_omit=omit,
                             report_include=include,
                             html_dir=directory,
                             extra_css=extra_css,
                             html_title=title,
                             html_skip_covered=skip_covered,
                             show_contexts=show_contexts,
                             report_contexts=contexts,
                             html_skip_empty=skip_empty,
                             precision=precision):
            reporter = HtmlReporter(self)
            return reporter.report(morfs)



import os
import subprocess
import sys

from mys.transpiler import TranspilerError

from .utils import Path
from .utils import TestCase
from .utils import create_new_package_with_files
from .utils import remove_ansi
from .utils import transpile_source


class Test(TestCase):

    def test_print(self):
        module_name = 'print'
        package_name = f'test_{module_name}'
        create_new_package_with_files(package_name, module_name, 'main')
        path = os.getcwd()

        with Path('tests/build/' + package_name):
            env = os.environ
            env['PYTHONPATH'] = path
            proc = subprocess.run([sys.executable, '-m', 'mys', 'run'],
                                  input="Lobster #1\n10\n",
                                  capture_output=True,
                                  text=True,
                                  env=env)

            if proc.returncode != 0:
                print(proc.stdout)
                print(proc.stderr)

                raise Exception("Build error.")

            output = remove_ansi(proc.stdout)

            self.assert_in(
                'A string literal!\n'
                '1\n'
                '1.5\n'
                'False\n'
                'True\n'
                'Foo(v=5)\n'
                '(-500, "Hi!")\n'
                '[1, 2, 3]\n'
                'Bar(a=Foo(v=3), b=True, c="kalle")\n'
                'Foo(v=5)\n'
                '[(Foo(v=3), True), (Foo(v=5), False)]\n'
                'True\n'
                'False\n'
                'True\n'
                'None\n'
                'Fie(a=5, _b=False, _c=None)\n'
                'G\n'
                '7\n'
                "['j', 'u', 'l']\n"
                'Fam(x=None)\n'
                'Fam(x=Foo(v=4))\n'
                'Fam(x=Bar(a=None, b=False, c="kk"))\n'
                'Foo(v=-1)\n'
                'Bar(a=None, b=True, c="")\n'
                '[Foo(v=5), Bar(a=None, b=True, c="fes")]\n'
                'b""\n'
                'b"\\x01\\x02\\x03"!\n'
                '20\n'
                '1\n' # Todo: Should print "Animal.Cow".
                'Name: Lobster #1\n'
                'Age: 10\n'
                'MyError(a=True, b="g")\n'
                '[1, 4, 3]\n'
                'END\n',
                output)
            self.assertTrue(('{1: 2, 3: 4}\n' in output)
                            or ('{3: 4, 1: 2}\n' in output))
            self.assertTrue(('{"ho": Foo(v=4), "hi": Foo(v=5)}\n' in output)
                            or ('{"hi": Foo(v=5), "ho": Foo(v=4)}\n' in output))
            self.assertTrue(('{7: 49, 1: 1}\n' in output)
                            or ('{1: 1, 7: 49}\n' in output))

    def test_basic_print_function(self):
        source = transpile_source('def main():\n'
                                  '    print(1)\n',
                                  has_main=True)

        self.assert_in('std::cout << 1 << "\\n";', source)

    def test_print_function_with_flush_true(self):
        source = transpile_source('def main():\n'
                                  '    print(1, flush=True)\n',
                                  has_main=True)

        self.assert_in('    std::cout << 1 << "\\n";\n'
                       '    if (mys::Bool(true)) {\n'
                       '        std::cout << std::flush;\n'
                       '    }',
                       source)

    def test_print_function_with_flush_false(self):
        source = transpile_source('def main():\n'
                                  '    print(1, flush=False)\n',
                                  has_main=True)

        self.assert_in('std::cout << 1 << "\\n";', source)

    def test_print_function_with_and_and_flush(self):
        source = transpile_source('def main():\n'
                                  '    print(1, end="!!", flush=True)\n',
                                  has_main=True)

        self.assert_in('std::cout << std::flush;\n',
                       source)

    def test_print_function_i8_u8_as_integers_not_char(self):
        source = transpile_source('def main():\n'
                                  '    print(i8(-1), u8(1), u16(1))\n',
                                  has_main=True)

        self.assert_in(
            '    std::cout << (int)i8(-1) << " " << (unsigned)u8(1) '
            '<< " " << u16(1) << "\\n";\n',
            source)

    def test_print_function_invalid_keyword(self):
        with self.assertRaises(TranspilerError) as cm:
            transpile_source('def main():\n'
                             '    print("Hi!", foo=True)\n',
                             'src/mod.mys',
                             'pkg/mod.mys.hpp',
                             has_main=True)

        self.assert_exception_string(
            cm,
            '  File "src/mod.mys", line 2\n'
            '        print("Hi!", foo=True)\n'
            '        ^\n'
            "CompileError: invalid keyword argument 'foo' to print(), only "
            "'end' and 'flush' are allowed\n")



from ..parser import ast
from .utils import CompileError


class ReturnCheckerVisitor(ast.NodeVisitor):
    """Check that given function always returns or raises.

    """

    def visit_body(self, node):
        for item in node:
            if self.visit(item):
                return True

        return False

    def visit_FunctionDef(self, node):
        if self.visit_body(node.body):
            return

        raise CompileError('missing return or raise', node)

    def visit_Return(self, _node):
        return True

    def visit_Raise(self, _node):
        return True

    def visit_If(self, node):
        if not self.visit_body(node.body):
            return False

        if node.orelse is None:
            return True
        else:
            return self.visit_body(node.orelse)

    def visit_Match(self, node):
        return all(self.visit_body(case.body) for case in node.cases)

    def visit_Try(self, node):
        if self.visit_body(node.finalbody):
            return True

        if not self.visit_body(node.body):
            return False

        return all(self.visit_body(handler.body) for handler in node.handlers)

    def visit_While(self, node):
        return WhileVisitor().visit(node)

class WhileVisitor(ast.NodeVisitor):
    """Returns True if the while loop always returns, False otherwise.

    """

    def visit_body(self, node):
        for item in node:
            if self.visit(item):
                return True

        return False

    def visit_While(self, node):
        if not isinstance(node.test, ast.Constant):
            return False

        if node.test.value is not True:
            return False

        return self.visit_body(node.body)

    def visit_For(self, _node):
        return False

    def visit_Break(self, _node):
        return False

    def visit_Return(self, _node):
        return True

    def visit_If(self, node):
        if self.visit_body(node.body):
            return True

        if node.orelse is None:
            return False
        else:
            return self.visit_body(node.orelse)

    def visit_Match(self, node):
        return any(self.visit_body(case.body) for case in node.cases)

    def visit_Try(self, node):
        if self.visit_body(node.finalbody):
            return True

        if not self.visit_body(node.body):
            return False

        return any(self.visit_body(handler.body) for handler in node.handlers)



# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/nedbat/coveragepy/blob/master/NOTICE.txt

"""Control of and utilities for debugging."""

import contextlib
import os
import sys

# When debugging, it can be helpful to force some options, especially when
# debugging the configuration mechanisms you usually use to control debugging!
# This is a list of forced debugging options.
FORCED_DEBUG = []
FORCED_DEBUG_FILE = None


class DebugControl:
    """Control and output for debugging."""

    show_repr_attr = False      # For SimpleReprMixin

    def __init__(self, options, output):
        """Configure the options and output file for debugging."""
        self.options = list(options) + FORCED_DEBUG
        self.suppress_callers = False

        filters = []
        self.output = DebugOutputFile.get_one(
            output,
            show_process=self.should('process'),
            filters=filters,
        )
        self.raw_output = self.output.outfile

    def should(self, option):
        """Decide whether to output debug information in category `option`."""
        if option == "callers" and self.suppress_callers:
            return False
        return (option in self.options)

    @contextlib.contextmanager
    def without_callers(self):
        """A context manager to prevent call stacks from being logged."""
        old = self.suppress_callers
        self.suppress_callers = True
        try:
            yield
        finally:
            self.suppress_callers = old


class NoDebugging:
    """A replacement for DebugControl that will never try to do anything."""
    def should(self, option):               # pylint: disable=unused-argument
        """Should we write debug messages?  Never."""
        return False


class SimpleReprMixin:
    """A mixin implementing a simple __repr__."""
    simple_repr_ignore = ['simple_repr_ignore', '$coverage.object_id']


class DebugOutputFile:                              # pragma: debugging
    """A file-like object that includes pid and cwd information."""
    def __init__(self, outfile, show_process, filters):
        self.outfile = outfile
        self.show_process = show_process
        self.filters = list(filters)

    SYS_MOD_NAME = '$coverage.debug.DebugOutputFile.the_one'

    @classmethod
    def get_one(cls, fileobj=None, show_process=True, filters=(), interim=False):
        """Get a DebugOutputFile.

        If `fileobj` is provided, then a new DebugOutputFile is made with it.

        If `fileobj` isn't provided, then a file is chosen
        (COVERAGE_DEBUG_FILE, or stderr), and a process-wide singleton
        DebugOutputFile is made.

        `show_process` controls whether the debug file adds process-level
        information, and filters is a list of other message filters to apply.

        `filters` are the text filters to apply to the stream to annotate with
        pids, etc.

        If `interim` is true, then a future `get_one` can replace this one.

        """
        if fileobj is not None:
            # Make DebugOutputFile around the fileobj passed.
            return cls(fileobj, show_process, filters)

        # Because of the way igor.py deletes and re-imports modules,
        # this class can be defined more than once. But we really want
        # a process-wide singleton. So stash it in sys.modules instead of
        # on a class attribute. Yes, this is aggressively gross.
        the_one, is_interim = sys.modules.get(cls.SYS_MOD_NAME, (None, True))
        if the_one is None or is_interim:
            if fileobj is None:
                debug_file_name = os.environ.get("COVERAGE_DEBUG_FILE",
                                                 FORCED_DEBUG_FILE)
                if debug_file_name:
                    fileobj = open(debug_file_name, "a")
                else:
                    fileobj = sys.stderr
            the_one = cls(fileobj, show_process, filters)
            sys.modules[cls.SYS_MOD_NAME] = (the_one, interim)
        return the_one

    def write(self, _text):
        """Just like file.write, but filter through all our filters."""
        self.outfile.flush()

    def flush(self):
        """Flush our file."""
        self.outfile.flush()



from mys.parser import ast
from mys.transpiler.iterators import transform

from .utils import TestCase


def remove_whitespace_lines(text):
    lines = []

    for line in text.splitlines():
        line = line.rstrip()

        if line:
            lines.append(line)

    return '\n'.join(lines)

class Test(TestCase):

    def assert_ok(self, iterator_code, class_code):
        tree = ast.parse(iterator_code).body[0]
        tree = transform(tree)
        self.assertEqual(
            remove_whitespace_lines(ast.unparse(ast.fix_missing_locations(tree))),
            remove_whitespace_lines(class_code))

    def test_empty(self):
        self.assert_ok(
            '@iterator\n'
            'def foo() -> i64:\n'
            '    pass\n'
            '',
            'class Foo:\n'
            '    _state: i64\n'
            '\n'
            '    def __init__(self):\n'
            '        self._state = 0\n'
            '\n'
            '    def next(self) -> i64:\n'
            '        while True:\n'
            '            match self._state:\n'
            '                case 0:\n'
            '                    pass\n'
            '                    raise RuntimeError()')

    def test_three_yield(self):
        self.assert_ok(
            '@iterator\n'
            'def foo() -> i64:\n'
            '    yield 1\n'
            '    yield 3\n'
            '    yield 5\n'
            '',
            'class Foo:\n'
            '    _state: i64\n'
            '\n'
            '    def __init__(self):\n'
            '        self._state = 0\n'
            '\n'
            '    def next(self) -> i64:\n'
            '        while True:\n'
            '            match self._state:\n'
            '                case 0:\n'
            '                    self._state = 1\n'
            '\n'
            '                    return 1\n'
            '                case 1:\n'
            '                    self._state = 2\n'
            '\n'
            '                    return 3\n'
            '                case 2:\n'
            '                    self._state = 3\n'
            '\n'
            '                    return 5\n'
            '                case 3:\n'
            '                    raise RuntimeError()')

    def test_if(self):
        self.assert_ok(
            '@iterator\n'
            'def foo() -> i64:\n'
            '    if True:\n'
            '        print(1)\n'
            '        yield 1\n'
            '    else:\n'
            '        yield 2\n'
            '        print(2)\n'
            '',
            'class Foo:\n'
            '    _state: i64\n'
            '\n'
            '    def __init__(self):\n'
            '        self._state = 0\n'
            '\n'
            '    def next(self) -> i64:\n'
            '        while True:\n'
            '            match self._state:\n'
            '                case 2:\n'
            '                    self._state = 1\n'
            '                case 3:\n'
            '                    print(2)\n'
            '                    self._state = 1\n'
            '                case 0:\n'
            '                    if True:\n'
            '                        print(1)\n'
            '                        self._state = 2\n'
            '\n'
            '                        return 1\n'
            '                    else:\n'
            '                        self._state = 3\n'
            '\n'
            '                        return 2\n'
            '                case 1:\n'
            '                    raise RuntimeError()')

    def test_while(self):
        self.assert_ok(
            '@iterator\n'
            'def foo() -> i64:\n'
            '    i = 0\n'
            '\n'
            '    while i < 10:\n'
            '        yield i\n'
            '        i += 1\n'
            '',
            'class Foo:\n'
            '    _state: i64\n'
            '\n'
            '    def __init__(self):\n'
            '        self._state = 0\n'
            '\n'
            '    def next(self) -> i64:\n'
            '        while True:\n'
            '            match self._state:\n'
            '                case 0:\n'
            '                    i = 0\n'
            '                    self._state = 2\n'
            '                case 3:\n'
            '                    i += 1\n'
            '                    self._state = 2\n'
            '                case 2:\n'
            '                    if i < 10:\n'
            '                        self._state = 3\n'
            '\n'
            '                        return i\n'
            '                    else:\n'
            '                        self._state = 1\n'
            '                case 1:\n'
            '                    raise RuntimeError()')

    def test_while_with_continue_and_break(self):
        with self.assertRaises(AssertionError):
            self.assert_ok(
                '@iterator\n'
                'def foo() -> i64:\n'
                '    i = 0\n'
                '\n'
                '    while i < 10:\n'
                '        if i == 1:\n'
                '            i += 2\n'
                '\n'
                '            while True:\n'
                '                break\n'
                '\n'
                '            continue\n'
                '        elif i == 4:\n'
                '            break\n'
                '        else:\n'
                '            yield i\n'
                '        i += 1\n'
                '',
                'class Foo:\n'
                '    _state: i64\n'
                '\n'
                '    def __init__(self):\n'
                '        self._state = 0\n'
                '\n'
                '    def next(self) -> i64:\n'
                '        while True:\n'
                '            match self._state:\n'
                '                case 0:\n'
                '                    i = 0\n'
                '                    self._state = 2\n'
                '                case 3:\n'
                '                    i += 1\n'
                '                    self._state = 2\n'
                '                case 2:\n'
                '                    if i < 10:\n'
                '                        if i == 1:\n'
                '                            i += 2\n'
                '\n'
                '                            while True:\n'
                '                                break\n'
                '\n'
                '                            self._state = 2\n'
                '                        elif i == 4:\n'
                '                            self._state = 1\n'
                '                        else:\n'
                '                            self._state = 3\n'
                '\n'
                '                            return i\n'
                '                    else:\n'
                '                        self._state = 1\n'
                '                case 1:\n'
                '                    raise RuntimeError()')

    def test_mixed(self):
        self.assert_ok(
            '@iterator\n'
            'def foo() -> i64:\n'
            '    i = 0\n'
            '    while i < 10:\n'
            '        if i == 5:\n'
            '            yield i\n'
            '        else:\n'
            '            yield 2 * i\n'
            '\n'
            '        i += 1\n'
            '\n'
            '    if True:\n'
            '        print(1 + 2)\n'
            '    elif False:\n'
            '        i = 1\n'
            '        while i != 1:\n'
            '            yield 5 - i\n'
            '    else:\n'
            '        print(1 - 1)\n'
            '',
            'class Foo:\n'
            '    _state: i64\n'
            '\n'
            '    def __init__(self):\n'
            '        self._state = 0\n'
            '\n'
            '    def next(self) -> i64:\n'
            '        while True:\n'
            '            match self._state:\n'
            '                case 0:\n'
            '                    i = 0\n'
            '                    self._state = 2\n'
            '                case 4:\n'
            '                    self._state = 3\n'
            '                case 5:\n'
            '                    self._state = 3\n'
            '                case 3:\n'
            '                    i += 1\n'
            '                    self._state = 2\n'
            '                case 2:\n'
            '                    if i < 10:\n'
            '                        if i == 5:\n'
            '                            self._state = 4\n'
            '\n'
            '                            return i\n'
            '                        else:\n'
            '                            self._state = 5\n'
            '\n'
            '                            return 2 * i\n'
            '                    else:\n'
            '                        self._state = 1\n'
            '                case 10:\n'
            '                    self._state = 9\n'
            '                case 9:\n'
            '                    if i != 1:\n'
            '                        self._state = 10\n'
            '\n'
            '                        return 5 - i\n'
            '                    else:\n'
            '                        self._state = 8\n'
            '                case 8:\n'
            '                    self._state = 7\n'
            '                case 7:\n'
            '                    self._state = 6\n'
            '                case 1:\n'
            '                    if True:\n'
            '                        print(1 + 2)\n'
            '                        self._state = 6\n'
            '                    elif False:\n'
            '                        i = 1\n'
            '                        self._state = 9\n'
            '                    else:\n'
            '                        print(1 - 1)\n'
            '                        self._state = 7\n'
            '                case 6:\n'
            '                    raise RuntimeError()')

    def test_fibonaccis(self):
        self.assert_ok(
            '@iterator\n'
            'def fibonaccis(count: i64) -> (i64, i64):\n'
            '    curr = 0\n'
            '    next = 1\n'
            '    i = 0\n'
            '\n'
            '    while i < count:\n'
            '        yield (i, curr)\n'
            '        yield (i, curr)\n'
            '        temp = curr\n'
            '        curr = next\n'
            '        next += temp\n'
            '        i += 1\n'
            '',
            'class Fibonaccis:\n'
            '    _state: i64\n'
            '\n'
            '    def __init__(self, count: i64):\n'
            '        self._state = 0\n'
            '\n'
            '    def next(self) -> i64:\n'
            '        while True:\n'
            '            match self._state:\n'
            '                case 0:\n'
            '                    curr = 0\n'
            '                    next = 1\n'
            '                    i = 0\n'
            '                    self._state = 2\n'
            '                case 3:\n'
            '                    self._state = 4\n'
            '\n'
            '                    return (i, curr)\n'
            '                case 4:\n'
            '                    temp = curr\n'
            '                    curr = next\n'
            '                    next += temp\n'
            '                    i += 1\n'
            '                    self._state = 2\n'
            '                case 2:\n'
            '                    if i < count:\n'
            '                        self._state = 3\n'
            '\n'
            '                        return (i, curr)\n'
            '                    else:\n'
            '                        self._state = 1\n'
            '                case 1:\n'
            '                    raise RuntimeError()')



from mys.transpiler import Source
from mys.transpiler import TranspilerError
from mys.transpiler import transpile

from .utils import TestCase
from .utils import build_and_test_module
from .utils import transpile_early_header


class Test(TestCase):

    def test_traits(self):
        build_and_test_module('traits')

    def test_define_empty_trait(self):
        header = transpile_early_header('@trait\n'
                                        'class Foo:\n'
                                        '    pass\n')
        self.assert_in('class Foo : public mys::Object {\n'
                       'public:\n'
                       '    virtual ~Foo() {}\n'
                       '};\n',
                       header)

    def test_define_trait_with_member(self):
        self.assert_transpile_raises(
            '@trait\n'
            'class Foo:\n'
            '    a: i32\n',
            '  File "", line 3\n'
            '        a: i32\n'
            '        ^\n'
            "CompileError: traits cannot have members\n")

    def test_define_trait_with_parameter(self):
        self.assert_transpile_raises(
            '@trait(u32)\n'
            'class Foo:\n'
            '    pass\n',
            '  File "", line 1\n'
            '    @trait(u32)\n'
            '     ^\n'
            "CompileError: no parameters expected\n")

    def test_define_trait_with_same_name_twice(self):
        self.assert_transpile_raises(
            '@trait\n'
            'class Foo:\n'
            '    pass\n'
            '@trait\n'
            'class Foo:\n'
            '    pass\n',
            '  File "", line 5\n'
            '    class Foo:\n'
            '    ^\n'
            "CompileError: there is already a trait called 'Foo'\n")

    def test_define_trait_with_same_name_as_class(self):
        self.assert_transpile_raises(
            'class Foo:\n'
            '    pass\n'
            '@trait\n'
            'class Foo:\n'
            '    pass\n',
            '  File "", line 4\n'
            '    class Foo:\n'
            '    ^\n'
            "CompileError: there is already a class called 'Foo'\n")

    def test_trait_does_not_exist(self):
        self.assert_transpile_raises(
            'class Foo(Bar):\n'
            '    pass\n',
            '  File "", line 1\n'
            '    class Foo(Bar):\n'
            '              ^\n'
            "CompileError: trait does not exist\n")

    def test_wrong_function_parameter_type_trait(self):
        self.assert_transpile_raises(
            '@trait\n'
            'class Base:\n'
            '    pass\n'
            '@trait\n'
            'class WrongBase:\n'
            '    pass\n'
            'class Foo(Base):\n'
            '    pass\n'
            'def foo(a: WrongBase):\n'
            '    pass\n'
            'def bar():\n'
            '    foo(Foo())\n',
            '  File "", line 12\n'
            '        foo(Foo())\n'
            '            ^\n'
            "CompileError: 'foo.lib.Foo' does not implement trait "
            "'foo.lib.WrongBase'\n")

    def test_trait_init(self):
        self.assert_transpile_raises(
            '@trait\n'
            'class Foo:\n'
            '    def __init__(self):\n'
            '        pass\n',
            '  File "", line 3\n'
            '        def __init__(self):\n'
            '        ^\n'
            "CompileError: traits cannot have an __init__ method\n")

    def test_pure_trait_method_not_implemented_pass(self):
        self.assert_transpile_raises(
            '@trait\n'
            'class Base:\n'
            '    def foo(self):\n'
            '        pass\n'
            'class Foo(Base):\n'
            '    pass\n',
            '  File "", line 5\n'
            '    class Foo(Base):\n'
            '              ^\n'
            "CompileError: pure trait method 'foo' is not implemented\n")

    def test_pure_trait_method_not_implemented_many_pass(self):
        self.assert_transpile_raises(
            '@trait\n'
            'class Base:\n'
            '    def foo(self):\n'
            '        pass\n'
            '        pass\n'
            '        pass\n'
            'class Foo(Base):\n'
            '    pass\n',
            '  File "", line 7\n'
            '    class Foo(Base):\n'
            '              ^\n'
            "CompileError: pure trait method 'foo' is not implemented\n")

    def test_pure_trait_method_not_implemented_docstring(self):
        self.assert_transpile_raises(
            '@trait\n'
            'class Base:\n'
            '    def foo(self):\n'
            '        "Doc"\n'
            'class Foo(Base):\n'
            '    pass\n',
            '  File "", line 5\n'
            '    class Foo(Base):\n'
            '              ^\n'
            "CompileError: pure trait method 'foo' is not implemented\n")

    def test_pure_trait_method_not_implemented_docstring_and_pass(self):
        self.assert_transpile_raises(
            '@trait\n'
            'class Base:\n'
            '    def foo(self):\n'
            '        "Doc"\n'
            '        pass\n'
            'class Foo(Base):\n'
            '    pass\n',
            '  File "", line 6\n'
            '    class Foo(Base):\n'
            '              ^\n'
            "CompileError: pure trait method 'foo' is not implemented\n")

    def test_imported_pure_trait_method_not_implemented(self):
        with self.assertRaises(TranspilerError) as cm:
            transpile([
                Source('@trait\n'
                       'class Base:\n'
                       '    def foo(self):\n'
                       '        pass\n',
                       module='foo.lib'),
                Source('from foo import Base\n'
                       'class Foo(Base):\n'
                       '    pass\n')
            ])

        self.assert_exception_string(
            cm,
            '  File "", line 2\n'
            '    class Foo(Base):\n'
            '              ^\n'
            "CompileError: pure trait method 'foo' is not implemented\n")

    def test_trait_member_access_1(self):
        self.assert_transpile_raises(
            '@trait\n'
            'class Base:\n'
            '    pass\n'
            'def foo(v: Base):\n'
            '    v.a = 1\n',
            '  File "", line 5\n'
            '        v.a = 1\n'
            '        ^\n'
            "CompileError: 'foo.lib.Base' has no member 'a'\n")

    def test_trait_member_access_2(self):
        self.assert_transpile_raises(
            'class Foo:\n'
            '    def foo(self, v: bool):\n'
            '        pass\n'
            'def foo(v: Foo):\n'
            '    v.foo()\n',
            '  File "", line 5\n'
            '        v.foo()\n'
            '        ^\n'
            "CompileError: parameter 'v: bool' not given\n")

    def test_wrong_trait_method_parameter_type(self):
        self.assert_transpile_raises(
            '@trait\n'
            'class Foo:\n'
            '    def foo(self, v: bool):\n'
            '        pass\n'
            'def foo(v: Foo):\n'
            '    v.foo(b"")\n',
            '  File "", line 6\n'
            '        v.foo(b"")\n'
            '              ^\n'
            "CompileError: expected a 'bool', got a 'bytes'\n")

    def test_trait_pascal_case(self):
        self.assert_transpile_raises(
            '@trait\n'
            'class foo:\n'
            '    pass\n',
            '  File "", line 2\n'
            '    class foo:\n'
            '    ^\n'
            "CompileError: trait names must be pascal case\n")

    def test_trait_member(self):
        self.assert_transpile_raises(
            '@trait\n'
            'class Foo:\n'
            '    x: i32\n',
            '  File "", line 3\n'
            '        x: i32\n'
            '        ^\n'
            "CompileError: traits cannot have members\n")



from ..parser import ast
from .utils import CompileError


class BodyCheckVisitor(ast.NodeVisitor):
    """Raises an error if given function body contains invalid nodes.

    """

    def visit_Expr(self, node):
        if isinstance(node.value, ast.Name):
            raise CompileError("bare name", node)
        elif isinstance(node.value, ast.Compare):
            raise CompileError("bare comparision", node)
        elif isinstance(node.value, ast.BinOp):
            raise CompileError("bare binary operation", node)
        elif isinstance(node.value, ast.UnaryOp):
            raise CompileError("bare unary operation", node)
        elif isinstance(node.value, ast.Constant):
            if isinstance(node.value.value, str):
                # ToDo: embedded C++
                pass
            elif isinstance(node.value.value, int):
                raise CompileError("bare integer", node)
            if isinstance(node.value.value, float):
                raise CompileError("bare float", node)
